package examplos.zookeeper;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.zookeeper.CreateMode;
import org.apache.zookeeper.WatchedEvent;
import org.apache.zookeeper.Watcher;
import org.apache.zookeeper.ZooDefs;
import org.apache.zookeeper.ZooKeeper;
import org.apache.zookeeper.data.Stat;

import com.google.gson.Gson;

public class Zookeeper {

	private ZooKeeper zookeeper;
	private static final int TIMEOUT = 5000;
	private CountDownLatch connectedSignal = new CountDownLatch(1);

	final String servers;

	public Zookeeper(String servers) throws Exception {
		this.servers = servers;
	}

	//Obtém a ligação ao zookeeper
	private ZooKeeper getZooKeeper() throws Exception {
		while (zookeeper == null || !zookeeper.getState().equals(ZooKeeper.States.CONNECTED)) {
			connect();
			Thread.sleep(1000);
		}
		return zookeeper;
	}

	//Faz a ligação ao zookeeper
	private void connect() throws IOException, InterruptedException {
		Logger.getAnonymousLogger().log(Level.INFO, "Estabelecendo ligação ao zookeeper em: " + servers);
		zookeeper = new ZooKeeper(servers, TIMEOUT, new Watcher() {
			@Override
			public void process(WatchedEvent event) {
				if (event.getState().equals(Watcher.Event.KeeperState.SyncConnected)) {
					connectedSignal.countDown();
				}
			}
		});
		connectedSignal.await();
	}

	/*
	 * Grava um objecto num dado caminho no Zookeeper
	 */
	public void gravarValor(String caminho, Object valor) {
		try {
			int i = caminho.lastIndexOf('/');
			if (i < 0)
				throw new IllegalArgumentException("O caminho tem que conter /");

			// Obter o caminho pai, à esquerda de /
			String pai = caminho.substring(0, i);

			// Criar o pai, se não existir
			criarCaminho(pai);

			// Converter o objecto valor para json e os respectivos bytes
			byte[] dados = new Gson().toJson(valor).getBytes();

			criarOuGravarNodo(caminho, dados);
		} catch (Exception x) {
			throw new ZookeeperException( x );
		}
	}

	/*
	 * Obtém a lista de objectos (da classe T) filhos de um dado caminho.
	 */
	public <T> List<T> listarValores(String caminho, Class<T> classOf ) {
		try {
			List<T> res = new ArrayList<T>();
			
			//Itera sobre os nós que são filhos de caminho
			for (String str : getZooKeeper().getChildren(caminho, false)) {
				String filho = caminho + "/" + str;
				
				//Obtém os dados gravados no caminho filho
				byte[] data = getZooKeeper().getData(filho, false, new Stat());
				//Converte o json para uma instância da classe T
				res.add( new Gson().fromJson(new String(data), classOf));
			}
			return res;
		} catch (Exception x) {
			throw new ZookeeperException( x );
		}
	}

	
	/*
	 * Cria um dado caminho se não existir. Associa ao caminho os dados passados em argumento.
	 */
	private void criarOuGravarNodo(String nodo, byte[] dados) {
		try {
			//Testa se o caminho já existe.
			Stat stat = getZooKeeper().exists(nodo, false);
			if (stat == null)
				//Se não existe, cria o nó zookeeper com os dados
				getZooKeeper().create(nodo, dados, ZooDefs.Ids.OPEN_ACL_UNSAFE, CreateMode.PERSISTENT);
			else
				//Se existe, substitui os dados associados ao nó zookeeper
				getZooKeeper().setData(nodo, dados, -1);

		} catch (Exception x) {
			throw new ZookeeperException(x);
		}
	}
	
	/*
	 *  Cria um caminho composto por vários sub-caminhos se não existir.
	 */
	public void criarCaminho(String caminho) {
		try {
			String[] partes = caminho.split("/");
			StringBuilder sb = new StringBuilder();
			for (String str : partes) {
				if (str.length() > 0) {
					sb.append('/').append(str);
					criarOuGravarNodo(sb.toString(), ".raiz".getBytes());
				}
			}
		} catch (Exception x) {
			throw new ZookeeperException(x);
		}
	}

	static class ZookeeperException extends RuntimeException {
		private static final long serialVersionUID = 1L;

		ZookeeperException(Throwable t) {
			super.initCause(t);
		}
	}
}
